#!/usr/bin/env python3
from __future__ import annotations
from lmn import __version__
from lmn import logger
import sys
import argparse


def global_parser():
    from . import run, sync, nv
    commands = [run, sync, nv]

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version="{} - version {}".format(
            "LMN", __version__
        )
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        default=False
    )

    # Register subparsers
    subparsers = parser.add_subparsers()
    name2subparser = {}
    for cmd in commands:
        subp = subparsers.add_parser(
            cmd.name,
            parents=[cmd.parser],  # HACK: this registers cmd.parser as the sub parser (https://docs.python.org/3.9/library/argparse.html#argparse.ArgumentParser)
            description=cmd.description,
            help=cmd.description,
            add_help=False  # Avoid help collision
            # formatter_class=PdmFormatter,
        )

        # NOTE: I don't particularly like this, but I follow how PDM handles (sub)commands.
        # This registers cmd.handler function as args.handler and it will be called later.
        subp.set_defaults(handler=cmd.handler)
        name2subparser[cmd.name] = subp
    return parser


def core(args):
    # Ensure same behavior while testing and using the CLI
    args = args or sys.argv[1:]

    # Get parser and parse arguments
    parser = global_parser()
    parsed = parser.parse_args(args)

    from logging import INFO, DEBUG
    logger.setLevel(INFO)
    if parsed.verbose:
        logger.setLevel(DEBUG)

    # Load config and fuse it with parsed arguments
    from ._config_loader import load_config
    project, remote_conf, preset_conf = load_config(parsed.machine)
    parsed.handler(project, remote_conf, parsed, preset_conf)


def main(args: list[str] | None = None) -> None:
    """
    Entrypoint for the CLI.
    Arguments are only for test (it is always None if calling via CLI).
    """
    core(args)



#!/usr/bin/env python3
from __future__ import annotations
import os
from pathlib import Path
from typing import Iterator

def is_system_root(directory: Path):
    return directory == directory.parent

def is_home_dir(directory: Path):
    return directory.resolve() == Path.home().resolve()

def yield_parents(directory: Path, max_depth=None) -> Iterator[Path]:
    """generator returning the the current directory and ancestors one after another"""
    max_depth = 999 if max_depth is None else max_depth
    for _ in range(max_depth):
        if is_system_root(directory):
            break
        directory = directory.parent
        yield directory


def merge_nested_dict(a, b, path=None, conflict='use_b'):
    """Merge dictionary b into a (can be nested)

    Correspondence: https://stackoverflow.com/a/7205107/7057866
    """
    if path is None: path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge_nested_dict(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass # same leaf value
            else:
                if conflict == 'use_b':
                    a[key] = b[key]
                elif conflict == 'use_a':
                    pass
                else:
                    raise ValueError('Conflict at %s' % '.'.join(path + [str(key)]))
        else:
            a[key] = b[key]
    return a


def parse_config(project_root):
    """Parse lmn config (json file).

    It looks for config file in this order:
    1. {project_root}/.lmn.json5
    3. $HOME/.config/lmn.json5
    """
    import pyjson5 as json

    from os.path import expandvars, isfile
    from lmn import logger
    from lmn.const import global_config_paths, local_config_fnames

    def _maybe_load(path):
        if isfile(path):
            with open(path, 'r') as f:
                return json.load(f)
        return {}

    # TODO: Hmmm... a better way to write this config search algorithm?
    path_found = False
    for path in global_config_paths:
        path = expandvars(path)
        if isfile(path):
            path_found = True
            break

    if path_found:
        with open(path, 'r') as f:
            global_conf = json.load(f)
    else:
        logger.error('lmn global config file cannot be located. Please place the config file at ~/.config/lmn.json5')
        global_conf = {}
        exit(1)

    # Loop over possible local config file names
    for local_conf_fname in local_config_fnames:
        local_conf = _maybe_load(f'{project_root}/{local_conf_fname}')

        if local_conf:
            break

    merged_conf = merge_nested_dict(global_conf, local_conf)
    return merged_conf


def find_project_root():
    """Find a project root (which is rsync-ed with the remote server).

    It first goes up in the directory tree to find ".git" or ".lmn.json5" file.
    If not found, print warning and just use the current directory
    """
    from lmn import logger
    from lmn.const import local_config_fnames
    def is_proj_root(directory: Path):
        if (directory / '.git').is_dir():
            return True
        for local_conf_fname in local_config_fnames:
            if (directory / local_conf_fname).is_file():
                return True
        return False

    current_dir = Path(os.getcwd()).resolve()
    if is_proj_root(current_dir):
        return current_dir

    for directory in yield_parents(current_dir):
        if is_proj_root(directory):
            return directory

    logger.warning('.git directory or .lmn.json5 file not found in the ancestor directories.\n'
                   'Setting project root to current directory')

    if is_system_root(current_dir):
        logger.error("project root detected is the system root '/' you never want to rsync your entire disk.")
        exit(1)

    if is_home_dir(current_dir):
        logger.error("project root detected is home directory. You never want to rsync the entire home directory to a remote machine.")
        exit(1)

    return current_dir


TIMESTAMP_FORMAT = '%Y-%m-%d_%H%M%S-%f'
from datetime import datetime
def get_timestamp() -> str:
    return datetime.strftime(datetime.now(), TIMESTAMP_FORMAT)


def read_timestamp(time_str: str) -> datetime:
    return datetime.strptime(time_str, TIMESTAMP_FORMAT)


def wrap_shebang(command, shell='bash'):
    return f"#!/usr/bin/env {shell}\n{command}"


sacct_cmd = "sacct --starttime $(date -d '40 hours ago' +%D-%R) --endtime now --format JobID,JobName%-100,NodeList,Elapsed,State,ExitCode --parsable2"


def parse_sacct(sacct_output):
    lines = sacct_output.strip().split('\n')
    keys = lines[0].split('|')
    entries = [{key: entry for key, entry in zip(keys, line.split('|'))} for line in lines[1:]]
    # NOTE: For a batch job, sacct shows two entries for one job submission:
    # {'JobID': '8231686', 'JobName': 'lmn-iti-coped-chief-97', ... 'MaxRSS': ''},
    # {'JobID': '8231686.batch', 'JobName': 'batch', ... 'MaxRSS': '64844148K'}
    # We should merge these into one.
    # TEMP: Forget about MaxRSS, and just remove all entries with *.batch
    entries = [entry for entry in entries if not entry['JobID'].endswith('.batch')]
    return entries


def posixpath2str(obj):
    import pathlib
    if isinstance(obj, list):
        return [posixpath2str(e) for e in obj]
    elif isinstance(obj, dict):
        return {key: posixpath2str(val) for key, val in obj.items()}
    elif isinstance(obj, pathlib.Path):
        return str(obj)
    else:
        return obj


def defreeze_dict(frozen_dict: frozenset):
    return {key: val for key, val in frozen_dict}


def replace_lmn_envvars(query: str, lmnenvs: dict):
    """
    Replace LMN_* envvars (i.e., ${LMN_CODE_DIR}, ${LMN_OUTPUT_DIR}, ${LMN_MOUNT_DIR}) with actual path.

    Args:
      - query (string): input string to process
      - lmnenvs (dict): map from LMN_* envvars to actual paths

    """
    import re
    lmn_envvars = ['LMN_CODE_DIR', 'LMN_OUTPUT_DIR', 'LMN_MOUNT_DIR']

    # Backward compatibility
    lmn_envvars += ['RMX_CODE_DIR', 'RMX_OUTPUT_DIR', 'RMX_MOUNT_DIR']

    for original in lmn_envvars:
        # Match "${original}" or "$original"
        regex = r'{}'.format('(\$\{' + original + '\}' + '|\$' + original + ')')
        target = lmnenvs[original]
        query = re.sub(regex, str(target), str(query))
    return query


from os.path import expandvars
class LaunchLogManager:
    def __init__(self, path=expandvars('$HOME/.lmn/launched.jsonl')) -> None:
        self.path = path

    def _refresh(self):
        import pyjson5 as json
        from datetime import datetime
        # Read entries from the top (oldest) to bottom (newest)
        with open(self.path, 'r') as f:
            entries = f.read().strip().split('\n')
        cutoff_idx = -1
        for idx, entry in enumerate(entries):
            timestamp = json.loads(entry).get('timestamp')
            dt = (datetime.now() - read_timestamp(timestamp)).seconds
            if dt <= 60 * 60 * 30:
                cutoff_idx = idx
                break
        refreshed_entries = entries[cutoff_idx:]

        # Write in jsonl format
        # NOTE: Make sure to add \n in the end!!
        with open(self.path, 'w') as f:
            f.write('\n'.join(refreshed_entries) + '\n')

    def log(self, entry):
        import json
        # Prepare jsonification
        entry = posixpath2str(entry)
        with open(self.path, 'a') as f:
            f.write(json.dumps(entry) + '\n')

    def read(self):
        pass



#!/usr/bin/env python3
import os
from os.path import expandvars
from lmn.helpers import posixpath2str, replace_lmn_envvars
import invoke

from lmn import logger

LMN_DOCKER_ROOTDIR = '/lmn'

# NOTE: Should I have ssh-conf, slurm-conf and docker-conf separately??
# I guess RemoteConfig should ONLY store the info on how to login to the host?
# docker info and slurm info should really reside in project.
class RemoteConfig:
    """Represents a configuration to connect to a remote server.
    This is used by SimpleSSHClient.
    """

    from fabric import Connection
    def __init__(self, user, host, port=22, slurm_node=False) -> None:
        self.user = user
        self.host = host
        # self.port = port
        self.slurm_node = slurm_node

    @property
    def base_uri(self) -> str:
        return f'{self.user}@{self.host}'

    def get_connection(self) -> Connection:
        from fabric import Connection
        from fabric.config import Config
        config = Config()
        config.user = self.user
        conn = Connection(host=self.host, config=config, inline_ssh_env=True)
        return conn

    def get_dict(self):
        return {key: val for key, val in vars(self).items() if not (key.startswith('__') or callable(val))}


class SimpleSSHClient:
    """Given a remote config, this provides an interface to ssh into a remote machine.
    """
    def __init__(self, remote_conf: RemoteConfig) -> None:
        self.remote_conf = remote_conf
        self.conn = self.remote_conf.get_connection()

    def uri(self, path):
        return f'{self.remote_conf.base_uri}:{path}'

    def run(self, cmd, directory='$HOME', disown=False, hide=False, env=None, pty=False, dry_run=False):
        import re
        
        # TODO: Check if $HOME would work or not!!
        env = {} if env is None else env

        # Perform shell escaping for envvars
        # NOTE: Fabric seems to have an issue to handle envvar that contains spaces...
        # This is the issue of using "inline_ssh_env" that essentially sets envvars by putting bunch of export KEY=VAL before running shells.
        # The documentation clearly says developers need to handle shell escaping for non-trivial values.
        # TEMP: shell escaping only when env contains space
        import shlex
        env = {key: shlex.quote(str(val)) if " " in str(val) else str(val) for key, val in env.items()}

        if dry_run:
            logger.info('--- dry run ---')
            logger.info(f'cmd: {cmd}')
            logger.debug(locals())
        else:
            with self.conn.cd(directory):
                # promise = self.conn.run(cmd, asynchronous=True)
                if disown:
                    # NOTE: asynchronous=True --> disown=True
                    # asynchronous=True returns a Promise to which you can attach and listen to stdio.
                    # disown=True completely disowns the process.
                    self.conn.run(cmd, disown=True, hide=hide, env=env, pty=pty)
                    return

                # NOTE: if you use asynchronous=True, stdout/stderr does not show up
                # when you use it on slurm. I have no idea why, tho.
                logger.debug(f'ssh client env: {env}')
                try:
                    result = self.conn.run(cmd, asynchronous=False, hide=hide, env=env, pty=pty)
                except invoke.exceptions.UnexpectedExit as e:
                    logger.info(f'Caught an exception!!:\n{str(e)}')
                    import sys
                    sys.exit(1)

            return result

    def put(self, file_like, target_path=None):
        self.conn.put(file_like, str(target_path))

    def port_forward(self):
        raise NotImplementedError

    def x_forward(self):
        raise NotImplementedError



#!/usr/bin/env python3

global_config_paths = ['${HOME}/.config/lmn.json5']
global_config_paths += ['${HOME}/.config/rmx']  # Backward compat

local_config_fnames = ['.lmn.json5']
local_config_fnames += ['.rmx.config']  # Backward compat



#!/usr/bin/env python3
import unittest
from lmn.helpers import parse_config, merge_nested_dict

class TestMergeDicts(unittest.TestCase):
    def test_flat_dict(self):
        """Merging flat dictionaries"""
        a = {'apple': 1, 'orange': 2}
        b = {'pearl': 3, 'lemon': 4}
        merged = merge_nested_dict(a, b)
        self.assertDictEqual({**a, **b}, merged)

    def test_flat_dict_dup_a(self):
        a = {'apple': 1, 'orange': 2}
        b = {'apple': 3, 'lemon': 4}
        merged = merge_nested_dict(a, b, conflict='use_a')
        self.assertDictEqual({**a, 'lemon': 4}, merged)

    def test_flat_dict_dup_b(self):
        a = {'apple': 1, 'orange': 2}
        b = {'apple': 3, 'lemon': 4}
        merged = merge_nested_dict(a, b, conflict='use_b')
        self.assertDictEqual({**b, 'orange': 2}, merged)

    def test_nested_dict_dup(self):
        a = {
            'machines': {'slurm': 'hoge', 'birch': 'fuga'},
            'slurm-configs': {'two-cpus': 'yeah'},
            'docker-images': {'image0': 'foo'}
        }
        b = {
            'machines': {'elm': 'piyo'},
            'docker-images': {'image2': 'xxx'}
        }
        gold = {
            'machines': {'slurm': 'hoge', 'birch': 'fuga', 'elm': 'piyo'},
            'slurm-configs': {'two-cpus': 'yeah'},
            'docker-images': {'image0': 'foo', 'image2': 'xxx'}
        }
        merged = merge_nested_dict(a, b)
        self.assertDictEqual(gold, merged)

    def test_nested_dict_dup2(self):
        """Lists are not merged but simply overwritten!"""
        a = {
            'machines': {'slurm': ['hoge', 'fuga']},
            'slurm-configs': {'two-cpus': 'yeah'},
            'docker-images': {'image0': 'foo'}
        }
        b = {
            'machines': {'slurm': ['piyo']},
            'docker-images': {'image0': ['bar', 'piyo']}
        }
        gold = {
            'machines': {'slurm': ['piyo']},
            'slurm-configs': {'two-cpus': 'yeah'},
            'docker-images': {'image0': ['bar', 'piyo']}
        }
        merged = merge_nested_dict(a, b)
        self.assertDictEqual(gold, merged)


if __name__ == '__main__':
    unittest.main()



"""Load config file and fuse it with runtime options"""
from __future__ import annotations
from argparse import Namespace
import os
import pathlib
from pathlib import Path
from lmn import logger
from lmn.helpers import find_project_root, parse_config
from posixpath import expandvars
from dotenv import dotenv_values

from lmn.machine import RemoteConfig

DOCKER_ROOT_DIR = '/lmn'

# NOTE: I used to set it to /tmp/lmn, but that causes an issue:
# When multiple users use this directory, the one who used this first set the permission of /tmp/lmn
# to be his/hers, thus others trying to use it later cannot access it.
REMOTE_ROOT_DIR = '/tmp'

class Project:
    """Maintains the info specific to the local project"""
    def __init__(self, name, rootdir, outdir=None, exclude=None, startup: str = "", 
                 mount_dirs: dict | None = None, mount_from_host: dict | None = None,
                 env: dict | None = None) -> None:
        self.name = name
        self.rootdir = Path(rootdir)
        self.outdir = self.rootdir / ".output" if outdir is None else outdir
        self.exclude = exclude
        self.startup = startup
        self.env = env if env is not None else {}
        self.mount_dirs = mount_dirs if mount_dirs is not None else {}
        self.mount_from_host = mount_from_host if mount_from_host is not None else {}

        self._make_directories()

    def _make_directories(self):
        self.rootdir.mkdir(parents=True, exist_ok=True)
        self.outdir.mkdir(parents=True, exist_ok=True)

    def get_dict(self):
        return {key: val for key, val in vars(self).items() if not (key.startswith('__') or callable(val))}

    def __repr__(self):
        return repr(f'<Project {self.name}>')


class Machine:
    """Maintains machine configuration.
    - RemoteConfig (user, hostname, uri)
    """
    def __init__(self, remote_conf: RemoteConfig, lmndir: str | Path,
                 parsed_conf: dict,
                 startup: str = "",
                 env: dict | None = None) -> None:
        self.remote_conf = remote_conf
        self.lmndir = Path(lmndir)
        self.env = env if env is not None else {}
        self.startup = startup
        self.parsed_conf = parsed_conf

        # aliases
        self.user = remote_conf.user
        self.host = remote_conf.host
        self.base_uri = remote_conf.base_uri

    def uri(self, path) -> str:
        """Returns user@hostname:path"""
        return f'{self.remote_conf.base_uri}:{path}'
    
    def get_lmndirs(self, project_name: str) -> Namespace:
        rootdir = self.lmndir / project_name
        return Namespace(
            codedir=str(rootdir / 'code'),
            mountdir=str(rootdir / 'mount'),
            outdir=str(rootdir / 'output')
        )


def get_docker_lmndirs(lmndir: Path | str, project_name: str) -> Namespace:
        rootdir = Path(lmndir) / project_name
        return Namespace(
            codedir=str(rootdir / 'code'),
            mountdir=str(rootdir / 'mount'),
            outdir=str(rootdir / 'output')
        )


def load_config(machine_name: str):
    proj_rootdir = find_project_root()
    config = parse_config(proj_rootdir)

    if machine_name not in config['machines']:
        raise KeyError(
            f'Machine "{machine_name}" not found in the configuration. '
            f'Available machines are: {" ".join(config["machines"].keys())}'
        )

    pconf = config.get('project', {})

    if machine_name not in config['machines']:
        raise KeyError(f'The specified machine {machine_name} is not found in the configuration file.')
    mconf = config['machines'].get(machine_name)

    # Parse special config params
    preset_conf = {
        'slurm-configs': config.get('slurm-configs', {}),
        'docker-images': config.get('docker-images', {}),
    }

    name = pconf.get('name', proj_rootdir.stem)
    logger.info(f'Project name     : {name}')
    logger.info(f'Project directory: {proj_rootdir}')

    mount_dirs = pconf.get('mount', [])
    mount_from_host = pconf.get('mount_from_host', {})

    if 'mount' in mconf:
        mount_dirs = mconf.get('mount', [])
    if 'mount_from_host' in mconf:
        mount_from_host = mconf.get('mount_from_host', {})

    # Load extra env vars from .env.secret
    secret_env_path = (proj_rootdir / ".secret.env").resolve()
    if not secret_env_path.is_file():
        secret_env_path = (proj_rootdir / ".env.secret").resolve()

        # Just a friendly reminder
        if secret_env_path.is_file():
            logger.info('Reading from ".env.secret" file will be deprecated in the future. Please rename it to ".secret.env".')

    secret_env = dotenv_values(secret_env_path)
    if secret_env:
        logger.debug(f'Loaded the following envs from secret env file: {dict(secret_env)}')

    project_env = pconf.get('environment', {})
    project = Project(name,
                      proj_rootdir,
                      outdir=pconf.get('outdir'),
                      exclude=pconf.get('exclude', []),
                      startup=pconf.get('startup', ""),
                      env={**project_env, **secret_env},
                      mount_dirs=mount_dirs,
                      mount_from_host=mount_from_host)

    user, host = mconf['user'], mconf['host']
    remote_conf = RemoteConfig(user, host)

    machine = Machine(remote_conf,
                      parsed_conf=mconf,
                      lmndir=mconf.get('root_dir', f'{REMOTE_ROOT_DIR}/{remote_conf.user}/lmn'),
                      env=mconf.get('environment', {}))

    return project, machine, preset_conf



from ._version import __version__  # NOQA
import colorlog

# NOTE: Here's the default color for colorlog
# The default colors to use for the debug levels
# default_log_colors = {
#     "DEBUG": "white",
#     "INFO": "green",
#     "WARNING": "yellow",
#     "ERROR": "red",
#     "CRITICAL": "bold_red",
# }

handler = colorlog.StreamHandler()
handler.setFormatter(colorlog.ColoredFormatter(
    '%(log_color)s%(levelname)s:%(name)s: %(message)s',
    log_colors={
        'DEBUG':    'cyan',
        'INFO':     'green',
        'WARNING':  'yellow',
        'ERROR':    'red',
        'CRITICAL': 'bold_red',
    }
))


logger = colorlog.getLogger('lmn')
logger.addHandler(handler)



#!/usr/bin/env python3
from __future__ import annotations
import os
from os.path import expandvars
import docker
from lmn import logger

class DockerContainerConfig:
    def __init__(self, image, name, env: dict | None = None, remove=True, network='host', ipc_mode='host', mounts=None,
                 startup: str | None = None, tty=True, use_gpus=True, user_id: int = 0, group_id: int = 0, runtime='docker') -> None:
        """
        user_id (int | str): The format is either 2003 for user_id or 2003:4000 to specify user id and group id.
        """
        self.image = image
        self.name = name
        self.env = {} if env is None else env
        self.remove = remove
        self.network = network
        self.ipc_mode = ipc_mode
        self.mounts = [] if mounts is None else mounts
        self.startup = startup
        self.tty = tty
        self.device_requests = []
        self.runtime = runtime
        self.use_gpus = use_gpus
        self.user_id = user_id
        self.group_id = group_id

        if use_gpus:
            self.add_gpus()

    def __repr__(self):
        properties = {key: val for key, val in vars(self).items()}
        return repr(f'[DockerContainerConfig]\n{properties}')

    def add_gpus(self):
        """Add gpus to device_requests

        https://github.com/docker/docker-py/issues/2395#issuecomment-907243275
        """
        gpu = docker.types.DeviceRequest(
            count=-1,
            capabilities=[['gpu'], ['nvidia'], ['compute'], ['compat32'], ['graphics'], ['utility'], ['video'], ['display']]
        )

        # Does "in" work here???
        if gpu not in self.device_requests:
            self.device_requests.append(gpu)

    def add_mount(self, target, source, type='bind', read_only=False):
        """Add a mount point"""
        from docker.types import Mount
        self.mounts.append(Mount(target=target, source=source, type=type, read_only=read_only))

    def use_x_forward(self, target_home=expandvars('$HOME')):
        """Set up X11 forwarding

        target_home: path to the home directory (on the remote server)
        Corresponds to https://github.com/afdaniele/x-docker
        """
        import shutil
        import subprocess

        # If installed with brew, it's stored in /opt/X11/bin/xhost, and it's not automatically appended to PATH
        xhost_command = '/opt/X11/bin/xhost' if os.path.isfile('/opt/X11/bin/xhost') else 'xhost'
        if not shutil.which(xhost_command):
            raise OSError('xhost not found. If you are on macOS, you need to install XQuartz.')

        # If using XQuartz, the display variables looks
        # something like: /private/tmp/com.apple.launchd.jY5AhC1lFM/org.xquartz:0
        # which doesn't work (??)
        if 'xquartz' in os.environ.get('DISPLAY', ''):
            from lmn.utils import run_cmd
            ip = run_cmd('ifconfig en0 | grep inet | awk \'$1=="inet" {print $2}\'', get_output=True, shell=True)
            display = f'{ip}:0'
        else:
            display = os.environ.get('DISPLAY', ':0'),

        # TODO: In addition to this, there needs to be a X forwarding from the remote to local
        # paramiko should support that.

        logger.info(f'display: {display}')
        self.environment.update({
            'DISPLAY': display,
            'NVIDIA_VISIBLE_DEVICES': 'all',  # Only if you have nvidia gpus
            'NVIDIA_DRIVER_CAPABILITIES': 'all',  # Only if you have nvidia gpus
            'QT_X11_NO_MITSHM': '1'  # support QT
        })
        # self.runtime = 'nvidia'  # You may need it if docker version is old. This is deprecated now.

        # run `xhost +local:root`
        subprocess.check_call([xhost_command, "+local:root"])

        # mount x-server socket
        self.add_mount(target='/tmp/.X11-unix', source='/tmp/.X11-unix')

        # TODO: Somehow needs to obtain $HOME in the remote server!
        # some sort of hook that is executed right before docker.run should work.
        # NOTE: This assumes the user in docker container is "root"
        self.add_mount(target='/root/.Xauthority', source=f'{target_home}/.Xauthority')


class SlurmConfig:
    """Let's keep it minimal. Conceptually it's better to store SlurmConfig here,
    but that would make it harder to read.
    """
    def __init__(self, job_name: str, partition='cpu', constraint=None, cpus_per_task=1, time=None,  
                 output=None, error=None, dependency=None, nodelist=None, exclude=None, shell='bash', **kwargs) -> None:
        self.job_name = job_name
        self.partition = partition
        self.constraint = constraint
        self.exclude = exclude
        self.cpus_per_task = cpus_per_task
        self.time = time
        self.dependency = dependency
        self.output = output
        self.error = error
        self.nodelist = nodelist
        self.shell = shell


class SingularityConfig:
    def __init__(self, image, env: dict | None = None, network='host', ipc_mode='host', mounts=None) -> None:
        self.image = image
        self.name = name
        self.env = {} if env is None else env
        self.remove = remove
        self.network = network
        self.ipc_mode = ipc_mode
        self.mounts = [] if mounts is None else mounts
        self.startup = startup
        self.tty = tty
        self.device_requests = []
        self.runtime = runtime
        self.use_gpus = use_gpus



from argparse import ArgumentParser, Namespace
from lmn import logger
from lmn.cli._utils import rsync
from lmn.cli._config_loader import Project, Machine
from lmn.machine import SimpleSSHClient


RSYNC_DESTINATION_PATH = "/tmp/".rstrip('/')


def _get_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument(
        "machine",
        action="store",
        type=str,
        help="Machine",
    )
    parser.add_argument(
        "--verbose",
        default=False,
        action="store_true",
        help="Be verbose"
    )
    return parser

def _sync_code(project: Project, machine: Machine, dry_run: bool = False):
    # rsync_options = f"--rsync-path='mkdir -p {project.remote_dir} && mkdir -p {project.remote_outdir} && mkdir -p {project.remote_mountdir} && rsync'"

    # A trick to create directories right before performing rsync
    lmndirs = machine.get_lmndirs(project.name)
    rsync_options = f"--rsync-path='mkdir -p {lmndirs.codedir} && mkdir -p {lmndirs.outdir} && mkdir -p {lmndirs.mountdir} && rsync'"

    try:
        rsync(source_dir=project.rootdir, target_dir=machine.uri(lmndirs.codedir),
                        exclude=project.exclude, options=rsync_options, dry_run=dry_run, transfer_rootdir=False)

        # rsync the directories to mount
        for mount_dir in project.mount_dirs:
            rsync(source_dir=mount_dir, target_dir=machine.uri(lmndirs.mountdir),
                                exclude=project.exclude, dry_run=dry_run)
    except OSError:
        import traceback
        import sys
        print(traceback.format_exc(0), file=sys.stderr)
        sys.exit(1)



def _sync_output(project: Project, machine: Machine, dry_run: bool = False):
    # Rsync remote outdir with the local outdir.
    if project.outdir:
        try:
            lmndirs = machine.get_lmndirs(project.name)
            # Check if there's any output file (the first line is always 'total [num-files]')
            ssh_client = SimpleSSHClient(machine.remote_conf)

            result = ssh_client.run(f'ls -l {lmndirs.outdir} | grep -v "^total" | wc -l', hide=True)
            num_output_files = int(result.stdout)

            msg = f'{num_output_files} files are in the output directory'
            logger.info(msg)
            if num_output_files > 0:
                rsync(source_dir=machine.uri(lmndirs.outdir), target_dir=project.outdir,
                    dry_run=dry_run)
                logger.info(f'The output files are copied to {str(project.outdir)}')

        except OSError:
            # NOTE: Only show the last stack of traceback (If I remember corectly...)
            import traceback
            import sys
            print(traceback.format_exc(0), file=sys.stderr)
            sys.exit(1)
    else:
        logger.warning('project.outdir is set to None. Doing nothing here.')


def handler(project: Project, machine: Machine, parsed: Namespace, preset: dict):
    """Deploy the local repository and execute the command on a machine.

    1. get SSH connection to machine
    2. run rsync(project.local_dir, project.remote_dir)
    3. run machine.execute(parsed.remote_command)
    """
    logger.debug(f'handling command for {__file__}')
    logger.debug(f'parsed: {parsed}')

    _sync_code(project, machine, dry_run=parsed.dry_run)
    _sync_output(project, machine, dry_run=parsed.dry_run)


name = 'sync'
description = 'sync command'
parser = _get_parser()



from __future__ import annotations
from copy import deepcopy
import os
from pathlib import Path
from argparse import ArgumentParser
from argparse import Namespace
from lmn import logger
from lmn.helpers import find_project_root
from lmn.config import SlurmConfig
from lmn.helpers import replace_lmn_envvars
from lmn.cli._config_loader import Project, Machine
from lmn.machine import SimpleSSHClient

from lmn.runner import SlurmRunner
from .sync import _sync_output, _sync_code


def _get_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument(
        "machine",
        action="store",
        type=str,
        help="Machine",
    )
    parser.add_argument(
        "--verbose",
        default=False,
        action="store_true",
        help="Be verbose"
    )
    parser.add_argument(
        "--image",
        default=None,
        help="specify a docker image"
    )
    parser.add_argument(
        "--name",
        default=None,
        help="specify docker container name"
    )
    parser.add_argument(
        "--sconf",
        default=None,
        help="specify a slurm configuration to be used"
    )
    parser.add_argument(
        "--dconf",
        default=None,
        help="specify a docker configuration to be used"
    )
    parser.add_argument(
        "-m",
        "--mode",
        action="store",
        type=str,
        default=None,
        choices=["ssh", "docker", "slurm", "singularity", "slurm-sing", "sing-slurm"],
        help="What mode to run",
    )
    parser.add_argument(
        "-d",
        "--disown",
        action="store_true",
        help="Do not block to wait for the process to exit. stdout/stderr will not be shown with this option.",
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="When a job with the same name already exists, kill it and run the new one (only for Docker mode)",
    )
    parser.add_argument(
        "-X",
        "--x-forward",
        action="store_true",
        help="X11 forwarding",
    )
    parser.add_argument(
        "--no-sync",
        action="store_true",
        help="Do not perform rsync. This means your local files will not be synced with remote server.",
    )
    parser.add_argument(
        "--contain",
        action="store_true",
        help="With this flag, rsync will copy the project directory to a new unique location on remote, rather than the predetermined one.",
    )
    parser.add_argument(
        "-n",
        "--num-sequence",
        action="store",
        type=int,
        default=1,
        help="number of sequence in Slurm sequential jobs"
    )
    parser.add_argument(
        "--sweep",
        action="store",
        type=str,
        help="specify sweep range (e.g., --sweep 0-255) this changes the value of $LMN_RUN_SWEEP_IDX"
    )
    parser.add_argument(
        "remote_command",
        default=False,
        action="store",
        nargs="+",
        type=str,
        help="Command to execute in a remote machine.",
    )
    return parser

from ._config_loader import Machine


def parse_sweep_idx(sweep_str):
    # Parse input
    # format #0: 8 --> 8
    # format #1: 1-10 --> range(1, 10)
    # format #2: 1,2,7 --> [1, 2, 7]
    if '-' in sweep_str:
        # format #1
        begin, end = [int(val) for val in sweep_str.split('-')]
        assert begin < end
        sweep_ind = range(begin, end)
    elif ',' in sweep_str:
        sweep_ind = [int(e) for e in sweep_str.strip().split(',')]
    elif sweep_str.isnumeric():
        sweep_ind = [int(sweep_str)]
    else:
        raise KeyError("Format for --sweep option is not recognizable. Format examples: '1-10', '8', '1,2,7'.")

    return sweep_ind


def print_conf(mode: str, machine: Machine, image: str | None = None):
    output = f'Running with [{mode}] mode on [{machine.remote_conf.base_uri}]'
    if image is not None:
        output += f' with image: [{image}]'
    logger.info(output)

def handler(project: Project, machine: Machine, parsed: Namespace, preset: dict):
    """
    Args:
    - project (Project): stores project-specific configurations
    - machine (Machine): stores machine-specific configurations
    - preset (dict)    : stores preset configurations for slurm or docker images
    """
    logger.debug(f'handling command for {__file__}')
    logger.debug(f'parsed: {parsed}')

    # Runtime info
    curr_dir = Path(os.getcwd()).resolve()
    proj_rootdir = find_project_root()
    rel_workdir = curr_dir.relative_to(proj_rootdir)
    logger.debug(f'relative working dir: {rel_workdir}')  # cwd.relative_to(project_root)
    if isinstance(parsed.remote_command, list):
        cmd = ' '.join(parsed.remote_command)
    else:
        cmd = parsed.remote_command

    runtime_options = Namespace(dry_run=parsed.dry_run,
                                cmd=cmd,
                                rel_workdir=rel_workdir,
                                disown=parsed.disown,
                                name=parsed.name,
                                sweep=parsed.sweep,
                                num_sequence=parsed.num_sequence,
                                no_sync=parsed.no_sync,
                                sconf=parsed.sconf,
                                dconf=parsed.dconf,
                                force=parsed.force)

    # Sync code first
    if parsed.no_sync:
        logger.warning('--no-sync option is True, local files will not be synced.')

    if not parsed.no_sync:
        if parsed.contain:
            # Generate a unique path and set it to machine.lmndir
            # BUG: This generates the same hash every time!! This stack overflow answer is obviously wrong: https://stackoverflow.com/a/6048639/19913466
            # import hashlib
            # import time
            # hashlib.sha1().update(str(time.time()).encode("utf-8"))
            # _hash = hashlib.sha1().hexdigest()
            from lmn.helpers import get_timestamp
            _hash = get_timestamp()
            machine.lmndir = Path(f'{machine.lmndir}/{_hash}')
            runtime_options.name = _hash
            logger.warning(f'--contain flag is set.\n\tsetting the remote lmndir to {machine.lmndir}\n\tsetting jobs suffix to {_hash}')

        _sync_code(project, machine, runtime_options.dry_run)

    env = {**project.env, **machine.env}
    env_from_host = []
    lmndirs = machine.get_lmndirs(project.name)

    startup = ' && '.join([e for e in [project.startup, machine.startup] if e.strip()])

    # If parsed.mode is not set, try to read from the config file.
    mode = parsed.mode or machine.parsed_conf.get('mode')
    if mode is None:
        logger.warning('mode is not set. Setting it to SSH mode')
        mode = 'ssh'

    if mode == 'ssh':
        from lmn.runner import SSHRunner
        ssh_client = SimpleSSHClient(machine.remote_conf)
        ssh_runner = SSHRunner(ssh_client, lmndirs)
        print_conf(mode, machine)
        ssh_runner.exec(runtime_options.cmd,
                        runtime_options.rel_workdir,
                        startup=startup,
                        env=env,
                        dry_run=runtime_options.dry_run)

    elif mode == 'docker':
        from docker import DockerClient
        from lmn.runner import DockerRunner
        from lmn.config import DockerContainerConfig
        base_url = "ssh://" + machine.base_uri
        # client = DockerClient(base_url=base_url, use_ssh_client=True)
        client = DockerClient(base_url=base_url)  # dockerpty hangs with use_ssh_client=True

        # Specify job name
        name = f'{machine.user}-lmn-{project.name}'
        if runtime_options.name is not None:
            name = f'{name}--{runtime_options.name}'

        if runtime_options.dry_run:
            raise ValueError('dry run is not yet supported for Docker mode')

        from docker.types import Mount
        from lmn.cli._config_loader import DOCKER_ROOT_DIR, get_docker_lmndirs
        docker_lmndirs = get_docker_lmndirs(DOCKER_ROOT_DIR, project.name)

        if not runtime_options.no_sync:
            mounts = [Mount(target=docker_lmndirs.codedir, source=lmndirs.codedir, type='bind'),
                    Mount(target=docker_lmndirs.outdir, source=lmndirs.outdir, type='bind'),
                    Mount(target=docker_lmndirs.mountdir, source=lmndirs.mountdir, type='bind')]
        else:
            mounts = []
        mounts += [Mount(target=tgt, source=src, type='bind') for src, tgt in project.mount_from_host.items()]

        docker_runner = DockerRunner(client, docker_lmndirs)

        # Docker specific configurations
        docker_pconf = machine.parsed_conf.get('docker', {})
        image = parsed.image or docker_pconf.get('image')
        user_id = docker_pconf.get('user_id', 0)
        group_id = docker_pconf.get('group_id', 0)

        if 'mount_from_host' in docker_pconf:
            logger.warn('''
            `mount_from_host` configuration under `docker` in a config file will be ignored.\n
            Please place it under `project`.
            ''')


        if not isinstance(user_id, int):
            raise ValueError('user_id must be an integer')

        if not isinstance(group_id, int):
            raise ValueError('group_id must be an integer')


        if image is None:
            raise KeyError('docker image is not specified.')

        print_conf(mode, machine, image)
        if runtime_options.sweep:
            assert runtime_options.disown, "You must set -d option to use sweep functionality."
            sweep_ind = parse_sweep_idx(runtime_options.sweep)

            single_sweep = (len(sweep_ind) == 1)

            for sweep_idx in sweep_ind:
                _name = f'{name}-{sweep_idx}'
                logger.info(f'Launching sweep {sweep_idx}: {_name}')
                env.update({'LMN_RUN_SWEEP_IDX': sweep_idx})
                docker_conf = DockerContainerConfig(
                    image=image,
                    name=_name,
                    mounts=mounts,
                    startup=startup,
                    env=env,
                    user_id=user_id,
                    group_id=group_id
                )
                docker_runner.exec(runtime_options.cmd,
                                   runtime_options.rel_workdir,
                                   docker_conf,
                                   interactive=False,
                                   kill_existing_container=runtime_options.force,
                                   quiet=not single_sweep)
            # import time
            # logger.warning('Sleeping for 5 seconds to see if the container fails...')
            # logger.warning('You can safely exit anytime')
            # time.sleep(10)
        else:
            docker_conf = DockerContainerConfig(
                image=image,
                name=name,
                mounts=mounts,
                startup=startup,
                env=env,
                user_id=user_id,
                group_id=group_id,
            )
            docker_runner.exec(runtime_options.cmd,
                               runtime_options.rel_workdir,
                               docker_conf,
                               interactive=not runtime_options.disown,
                               kill_existing_container=runtime_options.force)


    elif mode in ['slurm', 'slurm-sing', 'sing-slurm']:
        # Slurm specific configurations
        from lmn.config import SlurmConfig
        import randomname
        import random
        if 'slurm' not in machine.parsed_conf:
            raise ValueError('Configuration must have an entry for "slurm" to use slurm mode.')

        # NOTE: slurm seems to be fine with duplicated name.
        proj_name_maxlen = 15
        rand_num = random.randint(0, 100)
        job_name = f'lmn-{project.name[:proj_name_maxlen]}-{randomname.get_name()}-{rand_num}'

        # Parse from slurm config options (aside from default)
        if parsed.sconf is not None:
            logger.debug('parsed.sconf is specified. Loading custom preset conf.')
            sconf = preset.get('slurm-configs', {}).get(parsed.sconf, {})
            if sconf is None:
                raise KeyError(f'configuration: {parsed.sconf} cannot be found in "slurm-configs".')

        else:
            sconf = machine.parsed_conf['slurm']
        slurm_conf = SlurmConfig(job_name, **sconf)

        # logger.info(f'slurm_conf: {machine.sconf}')
        ssh_client = SimpleSSHClient(machine.remote_conf)
        slurm_runner = SlurmRunner(ssh_client, lmndirs)
        run_opt = runtime_options

        # Specify job name
        name = f'{machine.user}-lmn-{project.name}'
        if runtime_options.name is not None:
            name = f'{name}--{run_opt.name}'
        slurm_conf.job_name = name

        if mode in ['slurm-sing', 'sing-slurm']:
            # Decorate run_opt.cmd for Singularity

            # sconf = SlurmConfig(job_name, **mconf['slurm'])
            if machine.parsed_conf is None or 'singularity' not in machine.parsed_conf:
                raise ValueError('Entry "singularity" not found in the config file.')

            image = machine.parsed_conf.get('singularity', {}).get('sif_file')
            overlay = machine.parsed_conf.get('singularity', {}).get('overlay')
            writable_tmpfs = machine.parsed_conf.get('singularity', {}).get('writable_tmpfs', False)

            # Overwrite lmn envvars.  Hmm I don't like this...
            from lmn.cli._config_loader import DOCKER_ROOT_DIR, get_docker_lmndirs
            sing_lmndirs = get_docker_lmndirs(DOCKER_ROOT_DIR, project.name)

            # TODO: Do we need this??
            env.update({
                'LMN_CODE_DIR': sing_lmndirs.codedir, 
                'LMN_MOUNT_DIR': sing_lmndirs.mountdir,
                'LMN_OUTPUT_DIR': sing_lmndirs.outdir
            })

            # Backward compat
            env.update({
                'RMX_CODE_DIR': sing_lmndirs.codedir,
                'RMX_MOUNT_DIR': sing_lmndirs.mountdir,
                'RMX_OUTPUT_DIR': sing_lmndirs.outdir
            })

            # NOTE: Without --containall, nvidia-smi command fails with "couldn't find libnvidia-ml.so library in your system."
            # NOTE: Without bash -c '{cmd}', if you put PYTHONPATH=/foo/bar, it fails with no such file or directory 'PYTHONPATH=/foo/bar'
            # TODO: Will the envvars be taken over to the internal shell (by this extra bash command)?
            # sing_cmd = "singularity run --nv --containall {options} {sif_file} bash -c '{cmd}'"
            sing_cmd = 'singularity run --nv --containall {options} {sif_file} bash -c -- "{cmd}"'
            options = []

            # Bind
            bind = '-B {source}:{target}'
            if not runtime_options.no_sync:
                options += [bind.format(target=sing_lmndirs.codedir, source=lmndirs.codedir),
                            bind.format(target=sing_lmndirs.outdir, source=lmndirs.outdir),
                            bind.format(target=sing_lmndirs.mountdir, source=lmndirs.mountdir)]
            options += [bind.format(target=tgt, source=src) for src, tgt in project.mount_from_host.items()]

            # Overlay
            if overlay:
                options += [f'--overlay {overlay}']

            if writable_tmpfs:
                # This often solves the following error:
                # OSError: [Errno 30] Read-only file system
                options += ['--writable-tmpfs']

            # Environment variables
            # TODO: Better to use --env-file option and read from a file
            # Escaping quotes and commas will be much easier in that way.
            options += ['--env ' + ','.join(f'{key}="{val}"' for key, val in env.items())]

            # NOTE: Since CUDA_VISIBLE_DEVICES is often comma-separated values (i.e., CUDA_VISIBLE_DEVICES=0,1),
            # and `singularity run --env FOO=BAR,HOGE=PIYO` considers comma to be a separator for envvars,
            # It fails without special handling.
            env_from_host = ['CUDA_VISIBLE_DEVICES']

            # Workdir
            _workdir = sing_lmndirs.codedir / runtime_options.rel_workdir
            options += [f'--pwd {_workdir}']

            # Overwrite command
            options = ' '.join(options)

            # Trying my best to escape quotations (https://stackoverflow.com/a/18886646/19913466)
            logger.debug(f'run_opt.cmd before escape: {run_opt.cmd}')
            escaped_cmd = run_opt.cmd.encode('unicode-escape').replace(b'"', b'\\"').decode('utf-8')
            logger.debug(f'run_opt.cmd after escape: {escaped_cmd}')

            run_opt.cmd = sing_cmd.format(options=options, sif_file=image, cmd=escaped_cmd)

        print_conf(mode, machine, image=image if mode in ['slurm-sing', 'sing-slurm'] else None)
        if run_opt.sweep:
            assert run_opt.disown, "You must set -d option to use sweep functionality."
            sweep_ind = parse_sweep_idx(run_opt.sweep)

            _slurm_conf = deepcopy(slurm_conf)
            for sweep_idx in sweep_ind:
                # NOTE: This special prefix "SINGULARITYENV_" is stripped and the rest is passed to singularity container,
                # even with --containall or --cleanenv !!
                # Example: (https://docs.sylabs.io/guides/3.1/user-guide/environment_and_metadata.html?highlight=environment%20variable)
                #     $ SINGULARITYENV_HELLO=world singularity exec centos7.img env | grep HELLO
                #     HELLO=world
                env.update({'SINGULARITYENV_LMN_RUN_SWEEP_IDX': sweep_idx})

                # Oftentimes, a user specifies $LMN_RUN_SWEEP_IDX as an argument to the command,
                # and that will be evaluated right before singularity launches
                env.update({'LMN_RUN_SWEEP_IDX': sweep_idx})

                _name = f'{slurm_conf.job_name}-{sweep_idx}'
                logger.info(f'Launching sweep {sweep_idx}: {_name}')
                _slurm_conf.job_name = _name
                slurm_runner.exec(run_opt.cmd, run_opt.rel_workdir, slurm_conf=_slurm_conf,
                                  startup=startup,
                                  interactive=False, num_sequence=run_opt.num_sequence,
                                  env=env, env_from_host=env_from_host, dry_run=run_opt.dry_run)
        else:
            slurm_runner.exec(run_opt.cmd, run_opt.rel_workdir, slurm_conf=slurm_conf,
                              startup=startup, interactive=not run_opt.disown, num_sequence=run_opt.num_sequence,
                              env=env, env_from_host=env_from_host, dry_run=run_opt.dry_run)
    else:
        raise ValueError(f'Unrecognized mode: {mode}')

    # Sync output files
    if not runtime_options.no_sync:
        _sync_output(project, machine, dry_run=parsed.dry_run)

name = 'run'
description = 'run command'
parser = _get_parser()



#!/usr/bin/env python3
import subprocess
from lmn import logger

def rsync(source_dir, target_dir, options='', exclude=None, dry_run=False, transfer_rootdir=True):
    """
    source_dir: hoge/fuga/source-dir/content-files
    target_dir: Hoge/Fuga/target-dir

    if transfer_rootdir is True:
      target_dir: Hoge/Fuga/target-dir/source-dir/content-files

    else:
      target_dir: Hoge/Fuga/target-dir/content-files
    """
    # TODO: replace with https://github.com/laktak/rsyncy (?)
    # ^ This one supports visualizing progress bar

    import shutil
    exclude = [] if exclude is None else exclude
    # make sure rsync is installed
    if shutil.which("rsync") is None:
        raise RuntimeError("rsync binary is not found.")
    # ---
    logger.info(f"Syncing code...")
    source_dir = str(source_dir).rstrip('/') + ('' if transfer_rootdir else '/')
    target_dir = str(target_dir).rstrip('/') + '/'

    exclude_str = ' '.join(f'--exclude \'{ex}\'' for ex in exclude)

    # cmd = f"rsync --info=progress2 --archive --compress {exclude_str} {options} {source_dir} {target_dir}"
    cmd = f"rsync --progress --stats --archive --compress {exclude_str} {options} {source_dir} {target_dir}"
    logger.debug(f'running command: {cmd}')

    if not dry_run:
        out = run_cmd(cmd, shell=True)
        logger.info("Sync finished!")

        if out.returncode != 0:
            raise OSError(f'The following rsync command failed:\n{out.args}\n\n{out.stderr.decode("utf-8")}')
        return out


def run_cmd(cmd, get_output=False, shell=False) -> subprocess.CompletedProcess:

    if shell and isinstance(cmd, (list, tuple)):
        cmd = " ".join([str(s) for s in cmd])

    if get_output:
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=shell)
        proc.wait()
        if proc.returncode != 0:
            msg = "The command {} returned exit code {}".format(cmd, proc.returncode)
            raise RuntimeError(msg)
        out = proc.stdout.read().decode("utf-8").rstrip()
        logger.info(out)
        return out
    else:
        res = subprocess.run(cmd, shell=shell, capture_output=True)
        return res



from __future__ import annotations
from argparse import Namespace
from typing import Iterable, List, Optional
from io import StringIO
import threading
from lmn import logger
from docker import DockerClient
from lmn.config import DockerContainerConfig
from lmn.machine import SimpleSSHClient
from lmn.helpers import replace_lmn_envvars

def get_lmnenvs(cmd: str, lmndirs: Namespace):
    envvars = {
        "LMN_CODE_DIR": lmndirs.codedir,
        "LMN_MOUNT_DIR": lmndirs.mountdir,
        "LMN_OUTPUT_DIR": lmndirs.outdir,
        "LMN_USER_COMMAND": cmd,
    }

    # Provide RMX_* envvars for backward compatibility
    return {
        **envvars,
        **{key.replace("LMN", "RMX"): val for key, val in envvars.items()},
    }


class SSHRunner:
    def __init__(self, client: SimpleSSHClient, lmndirs: Namespace) -> None:
        self.client = client
        self.lmndirs = lmndirs

    def exec(self, cmd: str, relative_workdir, env: dict | None = None, startup: str = "", dry_run: bool = False,
             disown: bool = False):
        env = {} if env is None else env
        # if isinstance(cmd, list):
        #     cmd = ' '.join(cmd) if len(cmd) > 1 else cmd[0]

        if startup:
            cmd = f'{startup} && {cmd}'

        logger.debug(f'ssh run with command: {cmd}')
        logger.debug(f'cd to {self.lmndirs.codedir / relative_workdir}')
        lmnenv = get_lmnenvs(cmd, self.lmndirs)
        allenv = {**env, **lmnenv}
        allenv = {key: replace_lmn_envvars(val, lmnenv) for key, val in allenv.items()}
        return self.client.run(cmd, directory=(self.lmndirs.codedir / relative_workdir),
                               disown=disown, env=allenv, dry_run=dry_run, pty=True)


class DockerRunner:
    def __init__(self, client: DockerClient, lmndirs: Namespace) -> None:
        self.client = client
        self.lmndirs = lmndirs

    def exec(self, cmd: str, relative_workdir, docker_conf: DockerContainerConfig,
             kill_existing_container: bool = True, interactive: bool = True, quiet: bool = False,
             log_stderr_background: bool = False, use_cli: bool = True) -> None:
        if log_stderr_background:
            assert not interactive, 'log_stderr_background=True cannot be used with interactive=True'

        lmnenv = get_lmnenvs(cmd, self.lmndirs)
        allenv = {**docker_conf.env, **lmnenv}
        allenv = {key: replace_lmn_envvars(val, lmnenv) for key, val in allenv.items()}

        # NOTE: target: container, source: remote host
        logger.debug(f'mounts: {docker_conf.mounts}')
        logger.debug(f'docker_conf: {docker_conf}')


        logger.debug(f'container codedir: {str(self.lmndirs.codedir / relative_workdir)}')

        if kill_existing_container:
            from docker.errors import NotFound
            import docker
            try:
                container = self.client.containers.get(docker_conf.name)
            except NotFound:
                container = None
            
            if container:
                logger.warning(f'Removing the existing container: {docker_conf.name}')
                container.remove(force=True)

        # NOTE: Consider using Volume rather than Mount
        # - https://docker-py.readthedocs.io/en/1.2.3/volumes/
        # - https://docs.docker.com/storage/bind-mounts/

        # NOTE: Intentionally being super verbose to make arguments explicit.
        d = docker_conf

        # TEMP: When running in non-interactive mode and command fails, container disappears before we attach to its log stream,
        # and thus we cannot observe its error message. A naive way to avoid it is to wait for a bit before command execution.
        if not interactive:
            if docker_conf.startup:
                docker_conf.startup = ' && '.join((docker_conf.startup, 'sleep 2'))
            else:
                docker_conf.startup = 'sleep 2'

        if docker_conf.startup:
            cmd = ' && '.join((docker_conf.startup, cmd))
        cmd = f'{cmd} && chmod -R a+rw {str(self.lmndirs.outdir)}'

        if interactive:
            assert d.tty

            if use_cli:
                # Use python-on-whales (i.e., docker cli)
                import python_on_whales
                whale_client = python_on_whales.DockerClient(host=f'ssh://{self.client.api._custom_adapter.ssh_host}')
                logger.debug(f'docker run with command: {cmd}')

                # NOTE: dockerpy is stupid enough that it cannot attach remote pty.
                # dockerpty is good but fails when the terminal size changes frantically.
                # Let's wait for docker team to properly merge dockerpty.
                # Or I should probably think about migrating to https://github.com/gabrieldemarmiesse/python-on-whales
                # ^ This one just calls docker cli via Python, rather than using the complicated Docker SDK.
                # options = [
                #     '-it',
                #     '--gpus', 'all',
                #     '--workdir', str(self.lmndirs.codedir / relative_workdir),
                #     '--user', d.user_id,
                #     '--name', d.name,
                # ]
                # if d.network:
                #     options += ['--net', d.network]
                # if d.ipc_mode:
                #     options += ['--ipc', d.ipc_mode]
                # if allenv:
                #     options += [item for name, val in allenv.items() for item in ('-e', f"{name}='{val}'")]
                # if d.mounts:
                #     options += [item for m in d.mounts for item in ('-v', f'{m["Source"]}:{m["Target"]}')]

                # docker_cmd = [
                #     'docker',
                #     '-H', f'ssh://{self.client.api._custom_adapter.ssh_host}',
                #     'run',
                #     *options,
                #     d.image,
                #     cmd
                # ]
                # docker_cmd = [str(item) for item in docker_cmd]
                # logger.debug(f'running docker cli: {docker_cmd}')

                # import subprocess
                # subprocess.check_call(' '.join(docker_cmd), shell=True)

                try:
                    whale_client.run(
                        d.image,
                        ['/bin/bash', '-c', cmd],  # <-- TODO: THis needs to be a list of strings!!
                        detach=False,
                        envs=allenv,
                        gpus='all',
                        interactive=True,
                        ipc=d.ipc_mode,
                        mounts=[('type=bind', f'source={m["Source"]}', f'destination={m["Target"]}') for m in d.mounts],
                        name=d.name,
                        networks=[d.network],
                        remove=True,
                        tty=True,
                        user=f'{d.user_id}:{d.group_id}',
                        workdir=str(self.lmndirs.codedir / relative_workdir),
                    )
                except python_on_whales.exceptions.DockerException as e:
                    # NOTE: hide error as the exception is also raised when the command in the container returns non-zero exit value.
                    import traceback
                    import sys
                    logger.debug(f'python_on_whales failed!!:\n{str(e)}')
                    logger.debug(traceback.format_exc())


            else:
                # Use dockerpty
                cmd = f'/bin/bash -c \'{cmd}\''
                logger.debug(f'docker run with command: {cmd}')
                logger.debug('interactive is True. Force setting detach=False, tty=True, stdin_open=True.')
                container = self.client.containers.create(d.image,
                                                          cmd,
                                                          name=d.name,
                                                          network=d.network,
                                                          ipc_mode=d.ipc_mode,
                                                          detach=False,
                                                          tty=True,
                                                          stdin_open=True,
                                                          mounts=d.mounts,
                                                          environment=allenv,
                                                          device_requests=d.device_requests,
                                                          working_dir=str(self.lmndirs.codedir / relative_workdir),
                                                          user=f'{d.user_id}:{d.group_id}',
                                                          # entrypoint='/bin/bash -c "sleep 10 && xeyes"'  # Use it if you wanna overwrite entrypoint
                                                    )
                logger.debug(f'container: {container}')
                import dockerpty
                dockerpty.start(self.client.api, container.id)

        else:
            cmd = f'/bin/bash -c \'{cmd}\''
            container = self.client.containers.run(d.image,
                                                   cmd,
                                                   name=d.name,
                                                   remove=d.remove,  # Keep it running as we need to change
                                                   network=d.network,
                                                   ipc_mode=d.ipc_mode,
                                                   detach=True,
                                                   tty=(not log_stderr_background),  # If True, stdout/stderr are mixed, but I observed sometimes some stdout are not showing up when tty=False
                                                   stdin_open=True,  # It's useful to keep it open, as you may manually attach the container later
                                                   mounts=d.mounts,
                                                   environment=allenv,
                                                   device_requests=d.device_requests,
                                                   working_dir=str(self.lmndirs.codedir / relative_workdir),
                                                   user=f'{d.user_id}:{d.group_id}',
                                                )
            logger.debug(f'container: {container}')

            def log_stream(stream: Iterable):
                """print out log stream"""
                for char in stream:
                    # logger.info(char.decode('utf-8'))
                    # 'ignore' ignores decode error that happens when multi-byte char is passed.
                    print(char.decode('utf-8', 'ignore'), end='')


            if not quiet:
                if log_stderr_background:
                    # Attach log stream in a separate thread, and only output stderr
                    stream = container.logs(stdout=False, stderr=True, stream=True, follow=True)
                    logger.info('quiet is True, only listening to stderr.')
                    logger.info('--- listening container stderr ---\n')
                    thr = threading.Thread(target=log_stream, args=(stream, ))
                    thr.start()
                    # thr.join()  # This blocks forever ;P

                else:
                    # Block and listen to the stream from container
                    stream = container.logs(stream=True, follow=True)
                    logger.info('--- listening container stdout/stderr ---\n')
                    log_stream(stream)


class SlurmRunner:
    """Use srun/sbatch to submit the command on a remote machine.
    If your local machine has slurm (i.e., you're on slurm login-node), I guess you don't need this tool.
    Thus SlurmMachine inherits SSHMachine.
    """
    def __init__(self, client: SimpleSSHClient, lmndirs: Namespace) -> None:
        self.client = client
        self.lmndirs = lmndirs

    def exec(self, cmd: str, relative_workdir, slurm_conf, env: Optional[dict] = None,
             env_from_host: List[str] = [],
             startup: str = "", num_sequence: int = 1,
             interactive: bool = None, dry_run: bool = False):
        from simple_slurm_command import SlurmCommand
        env = {} if env is None else env

        # TODO: Verify env_from_list contains valid environment variable names
        # (i.e., cannot have spaces or quotes!!)

        if isinstance(cmd, list):
            cmd = ' '.join(cmd)

        if num_sequence > 1:
            # Use sbatch and set dependency to singleton
            slurm_conf.dependency = 'singleton'
            if interactive:
                logger.warning(f'num_sequence is set to {num_sequence} (> 1). Force disabling interactive mode')
                interactive = False

        s = slurm_conf
        slurm_command = SlurmCommand(cpus_per_task=s.cpus_per_task,
                                     job_name=s.job_name,
                                     partition=s.partition,
                                     time=s.time,
                                     nodelist=s.nodelist,
                                     exclude=s.exclude,
                                     constraint=s.constraint,
                                     dependency=s.dependency,
                                     output=s.output,
                                     error=s.error)

        if interactive and (s.output is not None):
            # User may expect stdout shown on the console.
            logger.info('--output/--error argument for Slurm is ignored in interactive mode.')

        lmnenv = get_lmnenvs(cmd, self.lmndirs)
        allenv = {**env, **lmnenv}
        allenv = {key: replace_lmn_envvars(val, lmnenv) for key, val in allenv.items()}

        if startup:
            cmd = f'{startup} && {cmd}'

        workdir = self.lmndirs.codedir / relative_workdir
        if interactive:
            # cmd = f'{shell} -i -c \'{cmd}\''
            # Create a temp bash file and put it on the remote server.
            exec_file = '\n'.join((
                f'#!/usr/bin/env {s.shell}',
                *[f'export SINGULARITYENV_{envvar}=${envvar}' for envvar in env_from_host],
                cmd
            ))
            logger.debug(f'exec file: {exec_file}')
            file_obj = StringIO(exec_file)
            self.client.put(file_obj, workdir / '.srun-script.sh')
            cmd = slurm_command.srun('.srun-script.sh', pty=s.shell)

            logger.debug(f'srun mode:\n{cmd}')
            logger.debug(f'cd to {workdir}')
            return self.client.run(cmd, directory=workdir,
                                   disown=False, env=allenv, pty=True, dry_run=dry_run)
        else:
            sbatch_cmd = slurm_command.sbatch(cmd, shell=f'/usr/bin/env {s.shell}')

            # HACK: rather than submitting with `sbatch << EOF\n ...\n EOF`,
            # I use `sbatch file-name` to avoid `$ENVVAR` to be evaluated right at the submission time
            # I want the `$ENVVAR` to be evaluated after the compute is allocated.
            sbatch_lines = sbatch_cmd.split('\n')[1:-1]  # Strip `sbatch << EOF` and `EOF`

            if env_from_host:
                # HACK: Inject `export` for exposing envvars to singularity container,
                # right after the last line that starts with `#SBATCH`.
                exports = [f'export SINGULARITYENV_{envvar}=${envvar}' for envvar in env_from_host]
                sbatch_lines = sbatch_lines[:-1] + exports + sbatch_lines[-1:]
            exec_file = '\n'.join(sbatch_lines)

            # TODO: If you're running a sweep, the content of the file should stay the same.
            # thus there shouldn't be a need to run these every time.
            file_obj = StringIO(exec_file)
            self.client.put(file_obj, workdir / '.sbatch-script.sh')

            logger.debug(f'sbatch mode\n===============\n{exec_file}\n===============')
            logger.debug(f'cd to {self.lmndirs.codedir / relative_workdir}')

            cmd = 'sbatch .sbatch-script.sh'
            cmd = '\n'.join([cmd] * num_sequence)
            result = self.client.run(cmd, directory=workdir,
                                     disown=False, env=allenv, dry_run=dry_run)
            if result.stderr:
                logger.warning('sbatch job submission failed:', result.stderr)
            jobid = result.stdout.strip().split()[-1]  # stdout format: Submitted batch job 8156833
            logger.debug(f'jobid {jobid}')
